import queue
import random
import threading
import time
from queue import Empty
from typing import List
from typing import Tuple

import requests

from .session import get_session


def _get_content(session: requests.Session, url: str, timeout: int):
    res = session.get(url, timeout=timeout)
    if res.status_code == 200:
        return res.content
    if res.status_code == 404:
        return b'Error, 404.'
    if res.status_code == 403:
        return b'Error, 403.'
    if res.status_code == 400:
        return b'Error, 400.'
    raise Exception(res.status_code)


def _loop(session: requests.Session,
          req_queue: queue.Queue,
          res: list,
          timeout: int, delay: int):
    if not session:
        session = get_session()

    while not req_queue.empty():
        time.sleep(delay * random.randint(1, 5))
        try:
            url = req_queue.get(block=False)
        except Empty:
            print(f"{threading.current_thread()}: finished.")
            break

        try:
            content = _get_content(session, url, timeout)
        except requests.exceptions.ConnectionError:
            print(f"{url}: Connection err.")
            res.append((url, b'Error, Connection err.'))
            req_queue.task_done()
            continue
        except Exception as e:
            print(f"{url}: Read time out or failure: {e}.")
            req_queue.put(url)
            req_queue.task_done()
            continue

        print(f"{threading.current_thread()}: finish {url}.")
        res.append((url, content))
        req_queue.task_done()


def mul_get_content(urls: List[str],
                    workers_num: int = 32,
                    delay: int = 0.5,
                    session: requests.Session = None,
                    timeout: int = 2) -> List[Tuple[str, bytes]]:
    """
    多线程发送GET请求，获取URL内容。
    常用于批量爬取页面，下载图片等用途。
    :param urls: 下载链接列表。
    :param workers_num: 开启线程数。
    :param delay: 等待基准时间 避免爬取过快 单位s.
    :param session: 是否使用统一的session。
    :param timeout: 默认超时等待时间。
    :return: (url, content)形式的列表。
    """
    req_queue = queue.Queue()
    for i in urls:
        req_queue.put(i)

    res = []
    for i in range(workers_num):
        th = threading.Thread(
            target=_loop,
            args=(session, req_queue, res, timeout, delay),
        )
        th.start()

    req_queue.join()
    return res



import json
import os
import sys

import tqdm

from licsber.utils.ufile import fun_check_path_exist, all_filepath, save_file
from licsber.utils.umeta import Meta
from licsber.utils.utime import cal_time


@cal_time(output=True)
@fun_check_path_exist(clean=True)
def save_115_link(start_path=None):
    if os.path.isfile(start_path):
        meta = Meta(start_path)
        print(meta)
        return

    res = ''
    waiting = list(all_filepath(start_path))
    waiting.sort()
    for filepath in tqdm.tqdm(waiting):
        meta = Meta(filepath)
        res += meta.gen_115_link() + '\n'

    save_file(start_path, '115_links.txt', res)


@cal_time(output=True)
@fun_check_path_exist(clean=True)
def save_115_dir(start_path=None):
    if not os.path.isdir(start_path):
        print(f"选定目录错误: {start_path}")
        return

    save_path = os.path.join(start_path, 'licsber_115.json')
    if os.path.exists(save_path):
        os.remove(save_path)

    def build(root):
        node = {
            'dir_name': os.path.basename(root),
            'dirs': [],
            'files': [],
        }
        for i in os.listdir(root):
            path = os.path.join(root, i)
            if os.path.isdir(path):
                node['dirs'].extend([build(path)])
            elif os.path.isfile(path):
                meta = Meta(path)
                link = meta.gen_115_link().lstrip('115://')
                node['files'].append(link)

        return node

    res = build(start_path)
    save_file(start_path, 'licsber_115.json', json.dumps(res))


@cal_time(output=True)
@fun_check_path_exist()
def conv(start_path=None):
    if not os.path.isfile(start_path):
        print(f"需要传入合法的115link文件: {start_path}")
        exit(-1)

    dirname = os.path.dirname(start_path)
    root_name = os.path.basename(dirname)
    root_name = sys.argv[2] if len(sys.argv) == 3 else root_name

    res, line = '', ''
    text = open(start_path).read().strip().split('\n')
    try:
        for line in text:
            filename, size, sha1, block = line.lstrip('115://').split('|')
            res += f"aliyunpan://{filename}|{sha1}|{size}|{root_name}\n"
    except ValueError:
        print(f"115Link不合法: {line}")
        exit(-1)

    save_file(dirname, 'aliyunpan_links.txt', res)



import base64

from licsber.github import get_secret
from pymongo import MongoClient
from pymongo.database import Database


class Mongo:
    def __init__(self, client, db_name, username, passwd):
        self.client = client
        self.db_name = db_name
        self.username = username
        self.passwd = passwd

    def connect(self) -> Database:
        db = self.client[self.db_name]
        db.authenticate(self.username, self.passwd)
        return db


def get_mongo(passwd_b64=get_secret('MONGO_PASSWD_B64'),
              host='mongodb://mongo.licsber.site',
              port=47107, db_name='auto', username='Automatic',
              connect=True):
    """
    获取mongo数据库连接, 用于爬虫.
    :param passwd_b64: base64后的密码.
    :param host: mongodb的url.
    :param port: host的端口.
    :param db_name: 数据库名称.
    :param username: 具有数据库权限的用户名.
    :param connect: 是否默认连接(选否会返回自定义对象 调用conn后才会认证).
    :return: mongo数据库连接.
    """
    if not passwd_b64:
        passwd_b64 = input('请输入mongo数据库连接密码(base64表示).')

    passwd = base64.b64decode(passwd_b64).decode('utf-8')
    c = MongoClient(host, port, connect=connect)
    if not connect:
        return Mongo(c, db_name, username, passwd)

    db = c[db_name]
    db.authenticate(username, passwd)
    return db



import numpy as np


# https://github.com/facebookresearch/Detectron/blob/master/detectron/utils/boxes.py
def xywh_to_xyxy(xywh):
    """Convert [x1 y1 w h] box format to [x1 y1 x2 y2] format."""
    if isinstance(xywh, (list, tuple)):
        # Single box given as a list of coordinates
        assert len(xywh) == 4
        x1, y1 = xywh[0], xywh[1]
        x2 = x1 + np.maximum(0., xywh[2] - 1.)
        y2 = y1 + np.maximum(0., xywh[3] - 1.)
        return (x1, y1, x2, y2)
    elif isinstance(xywh, np.ndarray):
        # Multiple boxes given as a 2D ndarray
        return np.hstack(
            (xywh[:, 0:2], xywh[:, 0:2] + np.maximum(0, xywh[:, 2:4] - 1))
        )
    else:
        raise TypeError('Argument xywh must be a list, tuple, or numpy array.')


def xyxy_to_xywh(xyxy):
    """Convert [x1 y1 x2 y2] box format to [x1 y1 w h] format."""
    if isinstance(xyxy, (list, tuple)):
        # Single box given as a list of coordinates
        assert len(xyxy) == 4
        x1, y1 = xyxy[0], xyxy[1]
        w = xyxy[2] - x1 + 1
        h = xyxy[3] - y1 + 1
        return (x1, y1, w, h)
    elif isinstance(xyxy, np.ndarray):
        # Multiple boxes given as a 2D ndarray
        return np.hstack((xyxy[:, 0:2], xyxy[:, 2:4] - xyxy[:, 0:2] + 1))
    else:
        raise TypeError('Argument xyxy must be a list, tuple, or numpy array.')



import os
import random
from collections import defaultdict

import cv2
import h5py
import numpy as np
from licsber.datasets import DATASETS_PATH


class SeTu1024:
    def __init__(self,
                 dataset_path=os.path.join(DATASETS_PATH, '1024'),
                 dataset_filename='1024.h5',
                 size=(224, 224)):
        if not os.path.exists(dataset_path):
            os.mkdir(dataset_path)

        self.path = dataset_path
        self.filename = dataset_filename

        self.label = {
            0: 'like',
            1: 'dislike',
        }
        self.items = {
            'like/train': 0,
            'like/test': 0,
            'dislike/train': 1,
            'dislike/test': 1,
        }
        self.size = size

    def make_dataset(self, test_ratio=0.1):
        random.seed(test_ratio)

        clean_ds_store(self.path)
        dir_paths = [os.path.join(self.path, i) for i in self.label.values()]
        for dir_path in dir_paths:
            dirs = ['train', 'test']
            for d in dirs:
                path = os.path.join(dir_path, d)
                if not os.path.exists(path):
                    os.mkdir(path)

            files = os.listdir(dir_path)
            files.remove('train')
            files.remove('test')

            random.shuffle(files)
            test_len = int(len(files) * test_ratio) + 1
            test = files[:test_len]
            for file in test:
                src = os.path.join(dir_path, file)
                dst = os.path.join(dir_path, 'test', file)
                os.rename(src, dst)

            for file in files[test_len:]:
                src = os.path.join(dir_path, file)
                dst = os.path.join(dir_path, 'train', file)
                os.rename(src, dst)

    def reset_dataset(self):
        dir_paths = [os.path.join(self.path, i) for i in self.label.values()]
        for dir_path in dir_paths:
            dirs = ['train', 'test']
            for d in dirs:
                path = os.path.join(dir_path, d)
                if os.path.exists(path):
                    for _, _, files in os.walk(path):
                        for file in files:
                            src = os.path.join(path, file)
                            dst = os.path.join(dir_path, file)
                            os.rename(src, dst)

            empty_dir(dir_path)

    @cal_time(output=True)
    def make_h5(self, save_path=None):
        if not save_path:
            save_path = os.path.join(self.path, self.filename)

        datasets = defaultdict(list)
        labels = defaultdict(list)

        for item in self.items:
            path = os.path.join(self.path, item)
            for img_filename in os.listdir(path):
                img_path = os.path.join(path, img_filename)
                img = cv2.imread(img_path)
                img = cv2.resize(img, self.size)
                label = [self.items[item]]
                if 'test' in item:
                    datasets['test'].append(img)
                    labels['test'].append(label)
                else:
                    datasets['train'].append(img)
                    labels['train'].append(label)

        with h5py.File(save_path, 'w') as h5:
            for d, l in zip(datasets, labels):
                h5.create_dataset(d, data=np.array(datasets[d]))
                h5.create_dataset(l + '_label', data=np.array(labels[l]))

    def load_dataset(self):
        h5_path = os.path.join(self.path, self.filename)
        if not os.path.exists(h5_path):
            print("请先生成数据集文件.")
            return None

        with h5py.File(h5_path) as h5:
            return (h5['train'][:], h5['train_label'][:]), (h5['test'][:], h5['test_label'][:])


def load_dataset():
    return SeTu1024().load_dataset()



import requests
from licsber.mongo import get_mongo
from licsber.utils.ubase64 import to_gbk_base64
from licsber.utils.utime import get_timestamp_str

URLS = {
    'set_user_bind': 'https://open.memobird.cn/home/setuserbind',
    'print_paper': 'https://open.memobird.cn/home/printpaper',
    'get_print_status': 'https://open.memobird.cn/home/getprintstatus'
}


def set_user_bind(ak: str, device_id: str, identifying='licsber') -> str:
    """
    绑定userid.
    :param ak: 开发者ak.
    :param device_id: 设备id 双击机器吐出来的.
    :param identifying: 用户自定义字符串.
    :return: 绑定成功的userid, 失败返回空串.
    """
    params = {
        'ak': ak,
        'timestamp': get_timestamp_str(),
        'memobirdID': device_id,
        'useridentifying': identifying,
    }
    res = requests.post(URLS['set_user_bind'], params=params).json()
    if res['showapi_res_code'] != 1:
        print(res)
        return ''
    return str(res['showapi_userid'])


def send_text_message(ak: str, device_id: str, text: str, userid=None) -> str:
    """
    向咕咕机发送纯文本消息.
    :param ak: 开发者ak.
    :param device_id: 设备id 双击机器吐出来的.
    :param text: 文本信息.
    :param userid: 可选的userid.
    :return: 打印id, 失败返回空串.
    """
    params = {
        'ak': ak,
        'timestamp': get_timestamp_str(),
        'printcontent': 'T:' + to_gbk_base64(text),
        'memobirdID': device_id,
        'userID': userid if userid else set_user_bind(ak, device_id, device_id),
    }
    res = requests.post(URLS['print_paper'], params=params).json()
    if res['showapi_res_code'] != 1:
        print(res)
        return ''
    return str(res['printcontentid'])


def get_status(ak: str, print_id: str) -> str:
    """
    获取消息发送的状态.
    :param ak: 开发者ak.
    :param print_id: 打印接口返回的打印id.
    :return: 已经打印返回1, 失败返回空串.
    """
    params = {
        'ak': ak,
        'timestamp': get_timestamp_str(),
        'printcontentid': print_id,
    }
    res = requests.post(URLS['get_print_status'], params=params).json()
    if res['showapi_res_code'] != 1:
        print(res)
        return ''
    return str(res['printflag'])


def log_message(mongo_passwd_b64: str, message: str):
    """
    生成mongo日志, 偷懒的接口.
    :param mongo_passwd_b64: mongo密码 同get_mongo的参数.
    :param message: 文本信息.
    :return: 插入结果.
    """
    db = get_mongo(mongo_passwd_b64)['memobird']
    return db.insert_one({
        'type': 'plain',
        'text': message,
        'print': False,
    })



from bs4 import BeautifulSoup as bs4
from licsber.auth import predict_captcha
from licsber.spider import get_session
from requests.compat import urljoin

from .wisedu_utils import need_captcha, check_captcha, get_captcha, encrypt


def get_wisedu_session(url, no, pwd, captcha_retry=5, remember_me=True):
    """
    对接"金智教务统一登录系统"的API.
    :param url: 访问跳转登录页的url, 通常以authserver开头.
    :param no: 学号.
    :param pwd: 密码.
    :param captcha_retry: 默认验证码重试次数(准确率90%以上).
    :param remember_me: 是否勾选免登录.
    :return: 一个登录完成的session, 可以继续访问接下来的网页.
    """

    def retry(session):
        require_captcha = need_captcha(url, session, no)
        captcha = ''
        if require_captcha:
            while not check_captcha(captcha):
                content = get_captcha(url, session)
                captcha = predict_captcha(content)

        res = session.get(url)
        data = {
            "lt": None,
            "dllt": None,
            "execution": None,
            "_eventId": None,
            "rmShown": None,
            'pwdDefaultEncryptSalt': None
        }

        res = bs4(res.content, 'html.parser')

        salt = res.find('input', id='pwdDefaultEncryptSalt')['value']
        login_url = res.find('form', id='casLoginForm')['action']
        login_url = urljoin(url, login_url)

        for i in res.find_all('input'):
            if 'name' in i.attrs and i['name'] in data:
                data[i['name']] = i['value']

        data['username'] = no
        data['password'] = encrypt(pwd, salt)
        if remember_me:
            data['rememberMe'] = 'on'
        if require_captcha:
            data['captchaResponse'] = captcha

        session.post(login_url, data=data)
        return len(session.cookies) != 2

    s = get_session()
    while captcha_retry and not retry(s):
        captcha_retry -= 1

    return s



import cv2
import matplotlib.pyplot as plt
import numpy as np


def parse_img(inputs):
    """
    将输入转化为图片, 支持路径/numpy array/content(图片二进制).
    :param inputs: 任意输入
    :return: numpy array
    """
    if type(inputs) == str:
        return cv2.imread(inputs)
    elif type(inputs) == bytes:
        return cv2.imdecode(np.frombuffer(inputs, np.uint8), cv2.COLOR_RGBA2RGB)
    else:
        return inputs


def imshow(img):
    """
    转换奇怪的bgr和rgb
    :param img: numpy array
    """
    if len(img.shape) != 3:
        plt.imshow(img)
        return

    b, g, r = cv2.split(img)
    rgb = cv2.merge([r, g, b])
    plt.imshow(rgb)
